@@ -426,6 +426,7 @@ emodules_load(const char *module, const char *modname, const char *modver)
     {
       emodules_depth--;
       dll_close (dlhandle);
+      dlhandle = 0;  /* Zero this out before module_load_unwind runs */
       return;
     }
 
@@ -581,7 +582,7 @@ when a dynamic module is loaded.
 */);
 
   /* #### Export this to Lisp */
-  Vmodule_extensions = build_string (":.ell:.so:.dll");
+  Vmodule_extensions = build_string (":.ell:.so:.dll:.dylib");
   staticpro (&Vmodule_extensions);
 
   load_modules_quietly = 0;
