@@ -108,6 +108,8 @@ Lisp_Object Vforward_word_regexp, Vbackward_word_regexp;
 Lisp_Object Vskip_chars_range_table;
 
 static void set_search_regs (struct buffer *buf, Bufpos beg, Charcount len);
+static void clear_unused_search_regs (struct re_registers *regp, int no_sub);
+/* #### according to comment in 21.5, unnecessary */
 static void save_search_regs (void);
 static Bufpos simple_search (struct buffer *buf, Bufbyte *base_pat,
 			     Bytecount len, Bytind pos, Bytind lim,
@@ -772,6 +774,7 @@ find_before_next_newline (struct buffer *buf, Bufpos from, Bufpos to, int count)
   return pos;
 }
 
+/* This function synched with FSF 21.1 */
 static Lisp_Object
 skip_chars (struct buffer *buf, int forwardp, int syntaxp,
 	    Lisp_Object string, Lisp_Object lim)
@@ -842,6 +845,7 @@ skip_chars (struct buffer *buf, int forwardp, int syntaxp,
 	    {
 	      Emchar cend;
 
+	      /* Skip over the dash.  */
 	      p++;
 	      if (p == pend) break;
 	      cend = charptr_emchar (p);
@@ -866,6 +870,7 @@ skip_chars (struct buffer *buf, int forwardp, int syntaxp,
 	}
     }
 
+  /* #### Not in FSF 21.1 */
   if (syntaxp && fastmap['-'] != 0)
     fastmap[' '] = 1;
 
@@ -879,36 +884,48 @@ skip_chars (struct buffer *buf, int forwardp, int syntaxp,
 
   {
     Bufpos start_point = BUF_PT (buf);
+    Bufpos pos = start_point;
+    Bytind pos_byte = BI_BUF_PT (buf);
 
     if (syntaxp)
       {
-	SETUP_SYNTAX_CACHE_FOR_BUFFER (buf, BUF_PT (buf), forwardp ? 1 : -1);
+	SETUP_SYNTAX_CACHE_FOR_BUFFER (buf, pos, forwardp ? 1 : -1);
 	/* All syntax designators are normal chars so nothing strange
 	   to worry about */
 	if (forwardp)
 	  {
-	    while (BUF_PT (buf) < limit
-		   && fastmap[(unsigned char)
-                              syntax_code_spec
-			      [(int) SYNTAX_FROM_CACHE (syntax_table,
-							BUF_FETCH_CHAR
-							(buf, BUF_PT (buf)))]])
-	      {
-		BUF_SET_PT (buf, BUF_PT (buf) + 1);
-		UPDATE_SYNTAX_CACHE_FORWARD (BUF_PT (buf));
-	      }
+	    if (pos < limit)
+	      while (fastmap[(unsigned char)
+			     syntax_code_spec
+			     [(int) SYNTAX_FROM_CACHE
+			      (syntax_table,
+			       BI_BUF_FETCH_CHAR (buf, pos_byte))]])
+		{
+		  pos++;
+		  INC_BYTIND (buf, pos_byte);
+		  if (pos >= limit)
+		    break;
+		  UPDATE_SYNTAX_CACHE_FORWARD (pos);
+		}
 	  }
 	else
 	  {
-	    while (BUF_PT (buf) > limit
-		   && fastmap[(unsigned char)
-                              syntax_code_spec
-			      [(int) SYNTAX_FROM_CACHE (syntax_table,
-							BUF_FETCH_CHAR
-							(buf, BUF_PT (buf) - 1))]])
+	    while (pos > limit)
 	      {
-		BUF_SET_PT (buf, BUF_PT (buf) - 1);
-		UPDATE_SYNTAX_CACHE_BACKWARD (BUF_PT (buf) - 1);
+		Bufpos savepos = pos_byte;
+		pos--;
+		DEC_BYTIND (buf, pos_byte);
+		UPDATE_SYNTAX_CACHE_BACKWARD (pos);
+		if (!fastmap[(unsigned char)
+			     syntax_code_spec
+			     [(int) SYNTAX_FROM_CACHE
+			      (syntax_table,
+			       BI_BUF_FETCH_CHAR (buf, pos_byte))]])
+		  {
+		    pos++;
+		    pos_byte = savepos;
+		    break;
+		  }
 	      }
 	  }
       }
@@ -916,36 +933,47 @@ skip_chars (struct buffer *buf, int forwardp, int syntaxp,
       {
 	if (forwardp)
 	  {
-	    while (BUF_PT (buf) < limit)
+	    while (pos < limit)
 	      {
-		Emchar ch = BUF_FETCH_CHAR (buf, BUF_PT (buf));
+		Emchar ch = BI_BUF_FETCH_CHAR (buf, pos_byte);
 		if ((ch < 0400) ? fastmap[ch] :
 		    (NILP (Fget_range_table (make_int (ch),
 					     Vskip_chars_range_table,
 					     Qnil))
 		     == negate))
-		  BUF_SET_PT (buf, BUF_PT (buf) + 1);
+		  {
+		    pos++;
+		    INC_BYTIND (buf, pos_byte);
+		  }
 		else
 		  break;
 	      }
 	  }
 	else
 	  {
-	    while (BUF_PT (buf) > limit)
+	    while (pos > limit)
 	      {
-		Emchar ch = BUF_FETCH_CHAR (buf, BUF_PT (buf) - 1);
+		Bufpos prev_pos_byte = pos_byte;
+		Emchar ch;
+
+		DEC_BYTIND (buf, prev_pos_byte);
+		ch = BI_BUF_FETCH_CHAR (buf, prev_pos_byte);
 		if ((ch < 0400) ? fastmap[ch] :
-		    (NILP (Fget_range_table (make_int (ch),
-					     Vskip_chars_range_table,
-					     Qnil))
-		     == negate))
-		  BUF_SET_PT (buf, BUF_PT (buf) - 1);
-                else
-                  break;
+		      (NILP (Fget_range_table (make_int (ch),
+					       Vskip_chars_range_table,
+					       Qnil))
+		       == negate))
+		  {
+		    pos--;
+		    pos_byte = prev_pos_byte;
+		  }
+		else
+		  break;
 	      }
 	  }
       }
     QUIT;
+    BOTH_BUF_SET_PT (buf, pos, pos_byte);
     return make_int (BUF_PT (buf) - start_point);
   }
 }
@@ -1146,10 +1174,11 @@ search_buffer (struct buffer *buf, Lisp_Object string, Bufpos bufpos,
   if (len == 0)
     {
       set_search_regs (buf, bufpos, 0);
+      clear_unused_search_regs (&search_regs, 0);
       return bufpos;
     }
 
-  /* Searching 0 times means don't move.  */
+  /* Searching 0 times means noop---don't move, don't touch registers.  */
   if (n == 0)
     return bufpos;
 
@@ -1196,6 +1225,8 @@ search_buffer (struct buffer *buf, Lisp_Object string, Bufpos bufpos,
 		    search_regs.start[i] += j;
 		    search_regs.end[i] += j;
 		  }
+	      /* re_match (called from re_search et al) does this for us */
+	      /* clear_unused_search_regs (search_regs, bufp->no_sub);   */
 	      XSETBUFFER (last_thing_searched, buf);
 	      /* Set pos to the new position. */
 	      pos = search_regs.start[0];
@@ -1233,6 +1264,8 @@ search_buffer (struct buffer *buf, Lisp_Object string, Bufpos bufpos,
 		    search_regs.start[i] += j;
 		    search_regs.end[i] += j;
 		  }
+	      /* re_match (called from re_search et al) does this for us */
+	      /* clear_unused_search_regs (search_regs, bufp->no_sub);   */
 	      XSETBUFFER (last_thing_searched, buf);
 	      /* Set pos to the new position. */
 	      pos = search_regs.end[0];
@@ -1433,6 +1466,7 @@ simple_search (struct buffer *buf, Bufbyte *base_pat, Bytecount len_byte,
 	  end = bytind_to_bufpos (buf, idx + buf_len);
 	}
       set_search_regs (buf, beg, end - beg);
+      clear_unused_search_regs (&search_regs, 0);
 
       return retval;
     }
@@ -1794,6 +1828,7 @@ boyer_moore (struct buffer *buf, Bufbyte *base_pat, Bytecount len,
 		    Bufpos bufend = bytind_to_bufpos (buf, bytstart + len);
 
 		    set_search_regs (buf, bufstart, bufend - bufstart);
+		    clear_unused_search_regs (&search_regs, 0);
 		  }
 
 		  if ((n -= direction) != 0)
@@ -1883,6 +1918,7 @@ boyer_moore (struct buffer *buf, Bufbyte *base_pat, Bytecount len,
 		    Bufpos bufend = bytind_to_bufpos (buf, bytstart + len);
 
 		    set_search_regs (buf, bufstart, bufend - bufstart);
+		    clear_unused_search_regs (&search_regs, 0);
 		  }
 
 		  if ((n -= direction) != 0)
@@ -1902,8 +1938,8 @@ boyer_moore (struct buffer *buf, Bufbyte *base_pat, Bytecount len,
   return bytind_to_bufpos (buf, pos);
 }
 
-/* Record beginning BEG and end BEG + LEN
-   for a match just found in the current buffer.  */
+/* Record the whole-match data (beginning BEG and end BEG + LEN) and the
+   buffer for a match just found.  */
 
 static void
 set_search_regs (struct buffer *buf, Bufpos beg, Charcount len)
@@ -1923,6 +1959,24 @@ set_search_regs (struct buffer *buf, Bufpos beg, Charcount len)
   XSETBUFFER (last_thing_searched, buf);
 }
 
+/* Clear unused search registers so match data will be null.
+   REGP is a pointer to the register structure to clear, usually the global
+   search_regs.
+   NO_SUB is the number of subexpressions to allow for.  (Does not count
+   the whole match, ie, for a string search NO_SUB == 0.)
+   It is an error if NO_SUB > REGP.num_regs - 1. */
+
+static void
+clear_unused_search_regs (struct re_registers *regp, int no_sub)
+{
+  /* This function has been Mule-ized. */
+  int i;
+
+  assert (no_sub >= 0 && no_sub < regp->num_regs);
+  for (i = no_sub + 1; i < regp->num_regs; i++)
+    regp->start[i] = regp->end[i] = -1;
+}
+
 
 /* Given a string of words separated by word delimiters,
    compute a regexp that matches those exact words
@@ -2825,8 +2879,11 @@ LIST should have been created by calling `match-data' previously.
   int num_regs;
   int length;
 
+#if 0
+  /* #### according to 21.5 comment, unnecessary */
   if (running_asynch_code)
     save_search_regs ();
+#endif
 
   CONCHECK_LIST (list);
 
@@ -2889,6 +2946,7 @@ LIST should have been created by calling `match-data' previously.
   return Qnil;
 }
 
+/* #### according to 21.5 comment, unnecessary */
 /* If non-zero the match data have been saved in saved_search_regs
    during the execution of a sentinel or filter. */
 static int search_regs_saved;
@@ -2912,6 +2970,8 @@ save_search_regs (void)
     }
 }
 
+/* #### according to 21.5 comment, unnecessary
+   prototype in lisp.h, all calls in process.c */
 /* Called upon exit from filters and sentinels. */
 void
 restore_match_data (void)
